/*
	Program name:	replace
	Author:			Andrew Marlow
	Date:			July 1989
*/

/* ------------------------------------------------------------------------ */
#include <stdio.h>
#include "portable.h"

#define IsAscii(c) (isascii(c))

#define MAX_LEN 256
#define MAX_NO_CHANGES 50

typedef struct filename_type
	{
		char Filename [256];
		struct filename_type *LeftPtr;
		struct filename_type *RightPtr;
	} FILENAME_TYPE;

static char String [MAX_LEN+1];
static char Search [MAX_NO_CHANGES] [MAX_LEN+1];
static char Replace [MAX_NO_CHANGES] [MAX_LEN+1];
static int NoOfChanges = 0;

/* ------------------------------------------------------------------------ */
EXTERN char *ErrorText ();
/* ------------------------------------------------------------------------ */
LOCAL VOID ProcessTree ();
LOCAL VOID ProcessFile ();
LOCAL FILENAME_TYPE *InsertItem ();
/* ------------------------------------------------------------------------ */


VOID main (argc, argv)
int argc;
char *argv[];
{
	int i, count, quit, MaxLen;
	FILENAME_TYPE *FilenamePtr;
	FILENAME_TYPE *RootPtr = NULL;
	
	MaxLen = 0;
	for (i = 1; i < argc; i++)
	{
		FilenamePtr = (FILENAME_TYPE *) malloc (sizeof(FILENAME_TYPE));
		if (FilenamePtr == NULL)
		{
			fprintf (stderr, "Error: malloc failed\n");
			exit (1);
		}
		strcpy (FilenamePtr->Filename, argv[i]);
		if (strlen (FilenamePtr->Filename) > MaxLen)
			MaxLen = strlen (FilenamePtr->Filename);
		FilenamePtr->LeftPtr = NULL;
		FilenamePtr->RightPtr = NULL;
		RootPtr = InsertItem (RootPtr, FilenamePtr);
	}

	count = argc-1;

	if (count == 0)
		printf ("There are no files to process.\n");
	else
	{
		quit = FALSE;
		NoOfChanges = 0;
		while (!quit)
		{
			printf ("Enter search string: ");
			gets (&Search[NoOfChanges][0]);
			if (strlen (&Search[NoOfChanges][0]) > 0)
			{
				printf ("Enter replace string: ");
				gets (&Replace[NoOfChanges][0]);
				NoOfChanges++;
			}
			else
				quit = TRUE;
		}
	}

	ProcessTree (RootPtr, count, MaxLen);
	exit (0);
}




LOCAL VOID ProcessTree (TreePtr, Count, MaxLen)
FILENAME_TYPE *TreePtr;
int Count;
int MaxLen;
{
	if (TreePtr == NULL)
		return;
		
	ProcessTree (TreePtr->LeftPtr, Count, MaxLen);
	ProcessFile (TreePtr->Filename, Count, MaxLen);
	ProcessTree (TreePtr->RightPtr, Count, MaxLen);
	free (TreePtr);
}


LOCAL VOID ProcessFile (InFilename, Count, MaxLen)
char *InFilename;
int Count;
int MaxLen;
{
	static int Number = 0;
	BOOLEAN BinaryFile, Changed;
	int i;
	FILE *InPtr, *OutPtr;
	char OutFilename[255];

	Number++;
	
	if ((InPtr = fopen (InFilename, "r")) == NULL)
	{
		fprintf (stderr, "Error: %s\nUnable to open '%s'\n", 
				 ErrorText(), InFilename);
		return;
	}

	strcpy (OutFilename, tmpnam(NULL));
	if ((OutPtr = fopen (OutFilename, "w")) == NULL)
	{
		fprintf (stderr, "Error: Unable to open '%s'\n", ErrorText(), OutFilename);
		return;
	}

	Changed = FALSE;

	printf ("%4d of %4d: %s ", Number, Count, InFilename);
	
	/* Would like to be able to use *%s feature of printf but some libraries 
	   have not implemented it or have bodged the implementation */
	   
	for (i = strlen(InFilename); i < MaxLen; i++)
		printf (" ");

	while (fgets (String, sizeof(String), InPtr) != NULL)
	{
		if (strlen(String) > 1)
			String[strlen(String)-1] = '\0';
		else
			String[0] = '\0';

		BinaryFile = FALSE;
		for (i = 0; i < strlen(String); i++)
			if (!IsAscii(String[i]))
			{
				BinaryFile = TRUE;
				break;
			}

		if (BinaryFile)
		{
			Changed = FALSE;
			break;
		}
		
		for (i = 0; i < NoOfChanges; i++)
			if (ReplaceString (String, &Search[i][0], &Replace[i][0]))
				Changed = TRUE;

		if (fprintf (OutPtr, "%s\n", String) != strlen(String)+1)
		{
			fprintf (stderr, "Unable to write to '%s'\n", OutFilename);
			return;
		}
	}
			
	fclose (InPtr);
	fclose (OutPtr);
		
	if (Changed)
	{
		if (DeleteFile (InFilename) == -1)
		{
			fprintf (stderr, "Unable to delete '%s'\n", InFilename);
			return;
		}
		if (RenameFile (OutFilename, InFilename) == -1)
		{
			fprintf (stderr, "Unable to rename '%s' to '%s'\n", OutFilename, InFilename);
			return;
		}
	}
	else
	{
		if (DeleteFile (OutFilename) == -1)
		{
			fprintf (stderr, "Unable to delete '%s'\n", OutFilename);
			return;
		}
	}

	printf ("%s\n", (Changed ? "***" : (BinaryFile ? "(binary ignored)" : "")));
}



LOCAL BOOLEAN ReplaceString (Source, Old, New)
char *Source;
char *Old;
char *New;
{
	BOOLEAN Changed = FALSE;
	int i, j, k;
	char Temp [MAX_LEN+1];

	j = k = 0;

	while ((i = strindex (&Source[j], Old)) != -1)
	{
		Changed = TRUE;
		strncpy (&Temp[k], &Source[j], i);
		strncpy (&Temp[k+i], New, strlen(New));
		j += i + strlen(Old);
		k += i + strlen(New);
	}
	strcpy (&Temp[k], &Source[j]);
	strcpy (Source, Temp);
	return (Changed);
}



LOCAL FILENAME_TYPE *InsertItem (RootPtr, ItemPtr)
FILENAME_TYPE *RootPtr;
FILENAME_TYPE *ItemPtr;
{
	if (RootPtr == NULL)
		return (ItemPtr);
	else
		if (strcmp (ItemPtr->Filename, RootPtr->Filename) < 0)
			RootPtr->LeftPtr = InsertItem (RootPtr->LeftPtr, ItemPtr);
		else
			RootPtr->RightPtr = InsertItem (RootPtr->RightPtr, ItemPtr);
	return (RootPtr);
}


LOCAL int strindex (s, t)
char *s, *t;
{
	int i, j, k;
	
	if (s == NULL)
		return (-1);

	for (i = 0; s[i] != '\0'; i++)
		if (strncmp (&s[i], t, strlen(t)) == 0)
			return (i);
	return (-1);
}
/* ------------------------------------------------------------------------ */

