/*
        Program name:   replace
        Author:         Andrew Marlow
        Date:           July 1989
*/

/* ------------------------------------------------------------------------ */
#include <stdio.h>

#include "portable.h"

#define IsAscii(c) (isascii(c))

#define MAX_LEN 256
#define MAX_NO_CHANGES 50

typedef struct filename_type
{
    char Filename[256];
    struct filename_type *LeftPtr;
    struct filename_type *RightPtr;
}   FILENAME_TYPE;

static char String[MAX_LEN + 1];
static char Search[MAX_NO_CHANGES][MAX_LEN + 1];
static char Replace[MAX_NO_CHANGES][MAX_LEN + 1];
static int NoOfChanges = 0;

/* ------------------------------------------------------------------------ */
GLOBAL char *ErrorText ();
/* ------------------------------------------------------------------------ */
LOCAL VOID DisplayUsage ();
LOCAL VOID ProcessTree ();
LOCAL VOID ProcessFile ();
LOCAL FILENAME_TYPE *InsertItem ();
LOCAL FILENAME_TYPE *AddItem ();
LOCAL FILENAME_TYPE *ProcessCatalog ();
LOCAL FILENAME_TYPE *ProcessDir ();
LOCAL int getline ();
LOCAL char *my_malloc ();
/* ------------------------------------------------------------------------ */



static int MaxLen;

VOID main (argc, argv)
    int argc;
    char *argv[];
{
    int i, count, quit;
    FILENAME_TYPE *FilenamePtr;
    FILENAME_TYPE *RootPtr = NULL;
    char cwd [100];

    MaxLen = 0;

    if (argc < 2)
    {
        DisplayUsage ();
        exit (1);
    }

    if (strcmp (argv[1], "-cat") == 0)
    {
        if (argc != 3)
        {
            DisplayUsage ();
            exit (1);
        }

        RootPtr = ProcessCatalog (RootPtr, argv[2], &count);
    }
    else if (strcmp (argv[1], "-dir") == 0 || strcmp (argv[1], "-d") == 0)
    {
        if (argc != 3)
        {
            DisplayUsage ();
            exit (1);
        }

        count = 0;
        getcwd (cwd, sizeof(cwd));
        if (chdir (argv[2]) == -1)
        {
            fprintf (stderr, "Error: Cannot change directory to '%s', %s\n", 
                     argv[2],  ErrorText());
            exit (1);
        }
        RootPtr = ProcessDir (RootPtr, argv[2], &count);
        chdir (cwd);
    }
    else
    {
        for (i = 1; i < argc; i++)
        {
            RootPtr = AddItem (RootPtr, argv[i]);
        }
        count = argc - 1;
    }

    if (count == 0)
        printf ("There are no files to process.\n");
    else
    {
        quit = FALSE;
        NoOfChanges = 0;
        while (!quit)
        {
            printf ("Enter search string: ");
            gets (&Search[NoOfChanges][0]);
            if (strlen (&Search[NoOfChanges][0]) > 0)
            {
                printf ("Enter replace string: ");
                gets (&Replace[NoOfChanges][0]);
                NoOfChanges++;
            }
            else
                quit = TRUE;
        }
    }

    ProcessTree (RootPtr, count, MaxLen);
    exit (0);
}



LOCAL VOID DisplayUsage ()
{
    fprintf (stderr, "Usage: repl <files>\n");
    fprintf (stderr, "            -cat <catalog>\n");
    fprintf (stderr, "            -dir <dir>\n");
}



LOCAL FILENAME_TYPE *ProcessCatalog (RootPtr, catalog, count)
FILENAME_TYPE *RootPtr;
char *catalog;
int *count;
{
    char buffer[1024];
    FILE *fptr;

    *count = 0;

    if ((fptr = fopen (catalog, "r")) == NULL)
    {
        fprintf (stderr, "Error: Cannot open '%s', %s\n", catalog,
                    ErrorText());
        exit (1);
    }

    while (getline (fptr, buffer, sizeof(buffer)) != -1)
    {
        if (strlen(buffer) > 0)
        {
            if (buffer[0] != '#')
            {
                RootPtr = AddItem (RootPtr, buffer);
                (*count)++;
            }
        }
    }

    fclose (fptr);
    return (RootPtr);
}



LOCAL FILENAME_TYPE *ProcessDir (RootPtr, dirname, count)
FILENAME_TYPE *RootPtr;
char *dirname;
int *count;
{
    DIR *dir_ptr = opendir (dirname);
    struct dirent *dir_info;
    char filename [1024];
    struct stat buf;

    if (dir_ptr == NULL)
    {
        fprintf (stderr, "Error: %s\nUnable to open '%s'\n",
                 ErrorText (), dirname);
        exit (1);
    }

    while ((dir_info = readdir (dir_ptr)) != NULL)
    {
        strncpy (filename, dir_info->d_name, dir_info->d_reclen);
        filename[dir_info->d_reclen] = '\0';

        /* If file type is directory then recurse */

        if (stat (filename, &buf) == -1)
        {
            fprintf (stderr, "Error: %s\nUnable to open '%s'\n",
                     ErrorText (), filename);
            return NULL;
        }

        if ((buf.st_mode & S_IFDIR) == S_IFDIR)
        {
            chdir (filename);
            RootPtr = ProcessDir (RootPtr, ".", count);
            chdir ("..");
        }
        else if ((buf.st_mode & S_IFREG) == S_IFREG)
        {
            RootPtr = AddItem (RootPtr, filename);
        }
    }

    closedir (dir_ptr);
    return (RootPtr);
}



LOCAL int getline (fp, buffer, buflen)
FILE *fp;
char *buffer;
int buflen;
{
    int c, i;

    i = 0;
    while (--buflen > 0 && (c = fgetc (fp)) != EOF && (c != '\n'))
        buffer[i++] = (char) c;
    buffer[i] = '\0';
    if (buflen > 0 && c == EOF)
    {
        return (-1);
    }
    for (i = strlen (buffer) - 1; i > 0 && buffer[i] == ' '; i--)
        buffer[i] = '\0';
    return (strlen (buffer));
}



LOCAL VOID ProcessTree (TreePtr, Count, MaxLen)
    FILENAME_TYPE *TreePtr;
    int Count;
    int MaxLen;
{
    if (TreePtr == NULL)
        return;

    ProcessTree (TreePtr->LeftPtr, Count, MaxLen);
    ProcessFile (TreePtr->Filename, Count, MaxLen);
    ProcessTree (TreePtr->RightPtr, Count, MaxLen);
    free (TreePtr);
}


LOCAL VOID ProcessFile (InFilename, Count, MaxLen)
    char *InFilename;
    int Count;
    int MaxLen;
{
    static int Number = 0;
    BOOLEAN BinaryFile, Changed, RegularFile, FileExists;
    int i;
    FILE *InPtr, *OutPtr;
    char OutFilename[255];
    struct stat buf;

    Number++;

    /* Check file type */

    if (stat (InFilename, &buf) == -1)
    {
        FileExists = FALSE;
    }
    else
        FileExists = TRUE;

    if (FileExists)
        RegularFile = ((buf.st_mode & S_IFREG) == S_IFREG) ? TRUE : FALSE;
    else
        RegularFile = FALSE;

    BinaryFile = FALSE;

    if (FileExists && RegularFile)
    {
        if ((InPtr = fopen (InFilename, "r")) == NULL)
        {
            fprintf (stderr, "Error: %s\nUnable to open '%s'\n",
                     ErrorText (), InFilename);
            return;
        }

        strcpy (OutFilename, tmpnam (NULL));
        if ((OutPtr = fopen (OutFilename, "w")) == NULL)
        {
            fprintf (stderr, "Error: Unable to open '%s'\n", ErrorText (), OutFilename);
            return;
        }
    }

    Changed = FALSE;

    printf ("%4d of %4d: %s ", Number, Count, InFilename);

    /*
     * Would like to be able to use *%s feature of printf but some libraries
     * have not implemented it or have bodged the implementation
     */

    for (i = strlen (InFilename); i < MaxLen; i++)
        printf (" ");
    fflush (stdout);

    if (FileExists && RegularFile)
    {
        while (fgets (String, sizeof (String), InPtr) != NULL)
        {
            if (strlen (String) > 1)
                String[strlen (String) - 1] = '\0';
            else
                String[0] = '\0';

            BinaryFile = FALSE;
            for (i = 0; i < strlen (String); i++)
                if (!isspace(String[i]) && !IsAscii (String[i]))
                {
                    BinaryFile = TRUE;
                    break;
                }
            if (BinaryFile)
            {
                Changed = FALSE;
                break;
            }

            for (i = 0; i < NoOfChanges; i++)
                if (ReplaceString (String, &Search[i][0], &Replace[i][0]))
                    Changed = TRUE;

            if (fprintf (OutPtr, "%s\n", String) != strlen (String) + 1)
            {
                fprintf (stderr, "Unable to write to '%s'\n", OutFilename);
                return;
            }
        }
        fclose (InPtr);
        fclose (OutPtr);

        if (Changed)
        {
            if (DeleteFile (InFilename) == -1)
            {
                fprintf (stderr, "Unable to delete '%s'\n", InFilename);
                return;
            }
            if (RenameFile (OutFilename, InFilename) == -1)
            {
                fprintf (stderr, "Unable to rename '%s' to '%s'\n", OutFilename, InFilename);
                return;
            }

            /* Change the permission back to what it was before */

            (void) chmod (InFilename, buf.st_mode);
        }
        else
        {
            if (DeleteFile (OutFilename) == -1)
            {
                fprintf (stderr, "Unable to delete '%s'\n", OutFilename);
                return;
            }
        }
    }

    if (Changed)
        printf ("***\n");
    else if (BinaryFile)
        printf ("(binary ignored)\n");
    else if (RegularFile)
        printf ("\n");
    else
    {
        if (FileExists)
            printf ("(file is not a regular file -- ignored)\n");
        else
            printf ("(file not found -- ignored)\n");
    }
}



LOCAL BOOLEAN ReplaceString (Source, Old, New)
    char *Source;
    char *Old;
    char *New;
{
    BOOLEAN Changed = FALSE;
    int i, j, k;
    char Temp[MAX_LEN + 1];

    j = k = 0;

    while ((i = strindex (&Source[j], Old)) != -1)
    {
        Changed = TRUE;
        strncpy (&Temp[k], &Source[j], i);
        strncpy (&Temp[k + i], New, strlen (New));
        j += i + strlen (Old);
        k += i + strlen (New);
    }
    strcpy (&Temp[k], &Source[j]);
    strcpy (Source, Temp);
    return (Changed);
}



LOCAL FILENAME_TYPE * AddItem (RootPtr, filename)
    FILENAME_TYPE *RootPtr;
    char *filename;
{
    FILENAME_TYPE *FilenamePtr;

    FilenamePtr = (FILENAME_TYPE *) my_malloc (sizeof (FILENAME_TYPE));
    strcpy (FilenamePtr->Filename, filename);
    if (strlen (FilenamePtr->Filename) > MaxLen)
        MaxLen = strlen (FilenamePtr->Filename);
    FilenamePtr->LeftPtr = NULL;
    FilenamePtr->RightPtr = NULL;
    RootPtr = InsertItem (RootPtr, FilenamePtr);
    return (RootPtr);
}



LOCAL FILENAME_TYPE * InsertItem (RootPtr, ItemPtr)
    FILENAME_TYPE *RootPtr;
    FILENAME_TYPE *ItemPtr;
{
    if (RootPtr == NULL)
        return (ItemPtr);
    else if (strcmp (ItemPtr->Filename, RootPtr->Filename) < 0)
        RootPtr->LeftPtr = InsertItem (RootPtr->LeftPtr, ItemPtr);
    else
        RootPtr->RightPtr = InsertItem (RootPtr->RightPtr, ItemPtr);
    return (RootPtr);
}

LOCAL char *my_malloc (nbytes)
unsigned nbytes;
{
    char *p;

    if (nbytes == 0)
    {
        fprintf (stderr, "Error: Attempt to malloc zero bytes.\n");
        exit (1);
    }
    p = malloc (nbytes);
    return (p);
}



LOCAL int strindex (s, t)
    char *s, *t;
{
    int i, j, k;

    if (s == NULL)
        return (-1);

    for (i = 0; s[i] != '\0'; i++)
        if (strncmp (&s[i], t, strlen (t)) == 0)
            return (i);
    return (-1);
}

/* ------------------------------------------------------------------------ */

